# sage_setup: distribution = sagemath-linbox

from sage.libs.flint.types cimport fmpz_t, fmpz_mat_t, fmpz_poly_t

# set C <- A * B
cdef void linbox_fmpz_mat_mul(fmpz_mat_t C, fmpz_mat_t A, fmpz_mat_t B) noexcept

# set cp to the characteristic polynomial of A
cdef void linbox_fmpz_mat_charpoly(fmpz_poly_t cp, fmpz_mat_t A) noexcept

# set mp to the minimal polynomial of A
cdef void linbox_fmpz_mat_minpoly(fmpz_poly_t mp, fmpz_mat_t A) noexcept

# return the rank of A
cdef size_t linbox_fmpz_mat_rank(fmpz_mat_t A) noexcept

# set det to the determinant of A
cdef void linbox_fmpz_mat_det(fmpz_t det, fmpz_mat_t A) noexcept



# sage_setup: distribution = sagemath-ntl
from sage.libs.ntl.types cimport ZZX_c

cdef class ntl_ZZX():
    cdef ZZX_c x
    cdef void setitem_from_int(ntl_ZZX self, long i, int value) noexcept
    cdef int getitem_as_int(ntl_ZZX self, long i) noexcept



# sage_setup: distribution = sagemath-ntl
from sage.libs.ntl.types cimport ZZ_c
from sage.libs.gmp.types cimport mpz_t, mpz_srcptr

cdef void ZZ_to_mpz(mpz_t output, ZZ_c* x) noexcept
cdef void mpz_to_ZZ(ZZ_c *output, mpz_srcptr x) noexcept
cdef void PyLong_to_ZZ(ZZ_c* z, value) noexcept



# sage_setup: distribution = sagemath-ntl
# distutils: depends = NTL/ZZ.h

from sage.libs.ntl.types cimport (ZZ_c, ZZ_p_c, ZZ_pContext_c, ZZ_pE_c, vec_ZZ_p_c,
        vec_ZZ_pE_c, ZZ_pEX_c, ZZ_pEX_Modulus_c)


cdef extern from "ntlwrap.h":
    long ZZ_pEX_IsZero "IsZero"(ZZ_pEX_c a)
    long ZZ_pEX_IsOne "IsOne"(ZZ_pEX_c a)

    void ZZ_pEX_add "add"( ZZ_pEX_c x, ZZ_pEX_c a, ZZ_pEX_c b)
    void ZZ_pEX_add_ZZ_p "add"( ZZ_pEX_c x, ZZ_pEX_c a, ZZ_p_c b)
    void ZZ_pEX_add_ZZ_pE "add"( ZZ_pEX_c x, ZZ_pEX_c a, ZZ_pE_c b)
    void ZZ_pEX_sub "sub"( ZZ_pEX_c x, ZZ_pEX_c a, ZZ_pEX_c b)
    void ZZ_pEX_sub_ZZ_p "sub"( ZZ_pEX_c x, ZZ_pEX_c a, ZZ_p_c b)
    void ZZ_pEX_sub_ZZ_pE "sub"( ZZ_pEX_c x, ZZ_pEX_c a, ZZ_pE_c b)
    void ZZ_pEX_negate "NTL::negate"(ZZ_pEX_c x, ZZ_pEX_c a)

    void ZZ_pEX_mul "mul"( ZZ_pEX_c x, ZZ_pEX_c a, ZZ_pEX_c b)
    void ZZ_pEX_mul_long "mul"( ZZ_pEX_c x, ZZ_pEX_c a, long b)
    void ZZ_pEX_mul_ZZ_p "mul"( ZZ_pEX_c x, ZZ_pEX_c a, ZZ_p_c b)
    void ZZ_pEX_mul_ZZ_pE "mul"( ZZ_pEX_c x, ZZ_pEX_c a, ZZ_pE_c b)
    void ZZ_pEX_sqr "sqr"( ZZ_pEX_c x, ZZ_pEX_c a)
    void ZZ_pEX_power "NTL::power"( ZZ_pEX_c x, ZZ_pEX_c a, long e)

    void ZZ_pEX_LeftShift "LeftShift"(ZZ_pEX_c x, ZZ_pEX_c a, long n)
    void ZZ_pEX_RightShift "RightShift"(ZZ_pEX_c x, ZZ_pEX_c a, long n)

    void ZZ_pEX_DivRem "DivRem"(ZZ_pEX_c q, ZZ_pEX_c r, ZZ_pEX_c a, ZZ_pEX_c b)
    void ZZ_pEX_div_ZZ_pEX "div"(ZZ_pEX_c q, ZZ_pEX_c a, ZZ_pEX_c b)
    void ZZ_pEX_div_ZZ_pE "div"(ZZ_pEX_c q, ZZ_pEX_c a, ZZ_pE_c b)
    void ZZ_pEX_div_ZZ_p "div"(ZZ_pEX_c q, ZZ_pEX_c a, ZZ_p_c b)
    void ZZ_pEX_div_long "div"( ZZ_pEX_c x, ZZ_pEX_c a, long b)
    void ZZ_pEX_rem "rem"(ZZ_pEX_c r, ZZ_pEX_c a, ZZ_pEX_c b)
    long ZZ_pEX_divide "divide"( ZZ_pEX_c x, ZZ_pEX_c a, ZZ_pEX_c b)

    void ZZ_pEX_GCD "GCD"(ZZ_pEX_c r, ZZ_pEX_c a, ZZ_pEX_c b)
    void ZZ_pEX_XGCD "XGCD"(ZZ_pEX_c d, ZZ_pEX_c s, ZZ_pEX_c t, ZZ_pEX_c a, ZZ_pEX_c b)

    long ZZ_pEX_deg "deg"( ZZ_pEX_c x )
    ZZ_pE_c ZZ_pEX_coeff "coeff"(ZZ_pEX_c a, long i)
    ZZ_pE_c ZZ_pEX_LeadCoeff "LeadCoeff"(ZZ_pEX_c a)
    void ZZ_pEX_SetCoeff "SetCoeff"(ZZ_pEX_c x, long i, ZZ_pE_c a)
    void ZZ_pEX_SetCoeff_ZZ_p "SetCoeff"(ZZ_pEX_c x, long i, ZZ_p_c a)
    void ZZ_pEX_SetCoeff_long "SetCoeff"(ZZ_pEX_c x, long i, long a)
    void ZZ_pEX_SetCoeff_one "SetCoeff"(ZZ_pEX_c x, long i)
    void ZZ_pEX_SetX "SetX"(ZZ_pEX_c x)
    long ZZ_pEX_IsX "IsX"(ZZ_pEX_c x)
    void ZZ_pEX_diff "diff"(ZZ_pEX_c x, ZZ_pEX_c a)
    void ZZ_pEX_MakeMonic "MakeMonic"(ZZ_pEX_c x)
    void ZZ_pEX_reverse_hi "reverse"(ZZ_pEX_c x, ZZ_pEX_c a, long hi)
    void ZZ_pEX_reverse "reverse"(ZZ_pEX_c x, ZZ_pEX_c a)

    void ZZ_pEX_random "random"(ZZ_pEX_c x, long n)

    void ZZ_pEX_BuildFromRoots "BuildFromRoots"(ZZ_pEX_c x, vec_ZZ_pE_c a)
    void ZZ_pEX_eval "eval"(ZZ_pE_c b, ZZ_pEX_c f, ZZ_pE_c a)
    void ZZ_pEX_eval_vec "eval"(vec_ZZ_pE_c b, ZZ_pEX_c f, vec_ZZ_pE_c a)
    void ZZ_pEX_interpolate "interpolate"(ZZ_pEX_c f, vec_ZZ_pE_c a, vec_ZZ_pE_c b)

    void ZZ_pEX_trunc "trunc"(ZZ_pEX_c x, ZZ_pEX_c a, long n)
    void ZZ_pEX_MulTrunc "MulTrunc"(ZZ_pEX_c x, ZZ_pEX_c a, ZZ_pEX_c b, long n)
    void ZZ_pEX_SqrTrunc "SqrTrunc"(ZZ_pEX_c x, ZZ_pEX_c a, long n)
    void ZZ_pEX_InvTrunc "InvTrunc"(ZZ_pEX_c x, ZZ_pEX_c a, long n)

    void ZZ_pEX_MulMod "MulMod"(ZZ_pEX_c x, ZZ_pEX_c a, ZZ_pEX_c b, ZZ_pEX_c f)
    void ZZ_pEX_SqrMod "SqrMod"(ZZ_pEX_c x, ZZ_pEX_c a, ZZ_pEX_c f)
    void ZZ_pEX_MulByXMod "MulByXMod"(ZZ_pEX_c x, ZZ_pEX_c a, ZZ_pEX_c f)
    void ZZ_pEX_InvMod "InvMod"(ZZ_pEX_c x, ZZ_pEX_c a, ZZ_pEX_c f)
    long ZZ_pEX_InvModStatus "InvModStatus"(ZZ_pEX_c x, ZZ_pEX_c a, ZZ_pEX_c f)

    void ZZ_pEX_Modulus_build "build"(ZZ_pEX_Modulus_c F, ZZ_pEX_c f)
    long ZZ_pEX_Modulus_deg "deg"(ZZ_pEX_Modulus_c F)

    void ZZ_pEX_MulMod_pre "MulMod"(ZZ_pEX_c x, ZZ_pEX_c a, ZZ_pEX_c b, ZZ_pEX_Modulus_c F)
    void ZZ_pEX_SqrMod_pre "SqrMod"(ZZ_pEX_c x, ZZ_pEX_c a, ZZ_pEX_Modulus_c F)
    void ZZ_pEX_PowerMod_pre "PowerMod"(ZZ_pEX_c x, ZZ_pEX_c a, long e, ZZ_pEX_Modulus_c F)
    void ZZ_pEX_PowerMod_ZZ_pre "PowerMod"(ZZ_pEX_c x, ZZ_pEX_c a, ZZ_c e, ZZ_pEX_Modulus_c F)
    void ZZ_pEX_PowerXMod_pre "PowerXMod"(ZZ_pEX_c x, long e, ZZ_pEX_Modulus_c F)
    void ZZ_pEX_PowerXMod_ZZ_pre "PowerXMod"(ZZ_pEX_c x, ZZ_c e, ZZ_pEX_Modulus_c F)
    void ZZ_pEX_rem_pre "rem"(ZZ_pEX_c x, ZZ_pEX_c a, ZZ_pEX_Modulus_c F)
    void ZZ_pEX_DivRem_pre "DivRem"(ZZ_pEX_c q, ZZ_pEX_c r, ZZ_pEX_c a, ZZ_pEX_Modulus_c F)
    void ZZ_pEX_div_pre "div"(ZZ_pEX_c q, ZZ_pEX_c a, ZZ_pEX_Modulus_c F)

    void ZZ_pEX_CompMod "CompMod"(ZZ_pEX_c x, ZZ_pEX_c f, ZZ_pEX_c g, ZZ_pEX_Modulus_c F)

    void ZZ_pEX_MinPolyMod "MinPolyMod"(ZZ_pEX_c h, ZZ_pEX_c g, ZZ_pEX_c f)
    void ZZ_pEX_MinPolyMod_pre "MinPolyMod"(ZZ_pEX_c h, ZZ_pEX_c g, ZZ_pEX_Modulus_c F)

    void ZZ_pEX_TraceMod "TraceMod"(ZZ_pE_c x, ZZ_pEX_c a, ZZ_pEX_c f)
    void ZZ_pEX_TraceMod_pre "TraceMod"(ZZ_pE_c x, ZZ_pEX_c a, ZZ_pEX_Modulus_c F)
    void ZZ_pEX_TraceVec "TraceVec"(vec_ZZ_pE_c x, ZZ_pEX_c f)
    void ZZ_pEX_NormMod "NormMod"(ZZ_pE_c x, ZZ_pEX_c a, ZZ_pEX_c f)
    void ZZ_pEX_resultant "resultant"(ZZ_pE_c x, ZZ_pEX_c a, ZZ_pEX_c b)

    void ZZ_pEX_clear "clear"(ZZ_pEX_c x)
    void ZZ_pEX_set "set"(ZZ_pEX_c x)

    long ZZ_pEX_IterIrredTest "IterIrredTest"(ZZ_pEX_c x)
    long ZZ_pEX_DetIrredTest "DetIrredTest"(ZZ_pEX_c x)
    long ZZ_pEX_ProbIrredTest "ProbIrredTest"(ZZ_pEX_c x, long iter)


cdef extern from "ntlwrap_impl.h":
    void ZZ_pEX_conv_modulus(ZZ_pEX_c fout, ZZ_pEX_c fin, ZZ_pContext_c c)



# sage_setup: distribution = sagemath-linbox
# sage.doctest: needs sage.libs.flint sage.libs.linbox

from sage.arith.misc import previous_prime
from sage.libs.flint.fmpq cimport fmpq_is_zero, fmpq_set_mpq, fmpq_canonicalise
from sage.libs.flint.fmpq_mat cimport fmpq_mat_entry_num, fmpq_mat_entry_den, fmpq_mat_entry
from sage.libs.flint.fmpz cimport fmpz_init, fmpz_clear, fmpz_set_mpz, fmpz_one, fmpz_get_mpz, fmpz_add, fmpz_mul, fmpz_sub, fmpz_mul_si, fmpz_mul_si, fmpz_mul_si, fmpz_divexact, fmpz_lcm
from sage.libs.gmp.types cimport mpz_t
from sage.matrix.constructor import matrix
from sage.matrix.matrix_cyclo_dense cimport Matrix_cyclo_dense
from sage.matrix.matrix_integer_linbox cimport _lift_crt
from sage.matrix.matrix_rational_dense cimport Matrix_rational_dense
from sage.matrix.matrix_space import MatrixSpace
from sage.matrix.misc_flint import matrix_integer_dense_rational_reconstruction
from sage.misc.verbose import verbose
from sage.rings.integer_ring import ZZ
from sage.rings.rational_field import QQ
from sage.structure.element cimport Matrix as baseMatrix
from sage.structure.proof.proof import get_flag as get_proof_flag

from sage.matrix.matrix_modn_dense_double import MAX_MODULUS as MAX_MODULUS_modn_dense_double
from sage.arith.multi_modular import MAX_MODULUS as MAX_MODULUS_multi_modular
MAX_MODULUS = min(MAX_MODULUS_modn_dense_double, MAX_MODULUS_multi_modular)

# parameters for tuning
echelon_primes_increment = 15
echelon_verbose_level = 1


cpdef _matrix_times_matrix_(Matrix_cyclo_dense self, baseMatrix right):
    """
    Return the product of two cyclotomic dense matrices.

    INPUT:

    - ``self``, ``right`` -- cyclotomic dense matrices with compatible
      parents (same base ring, and compatible dimensions for matrix
      multiplication)

    OUTPUT: cyclotomic dense matrix

    ALGORITHM:

    Use a multimodular algorithm that involves multiplying the two matrices
    modulo split primes.

    EXAMPLES::

        sage: W.<z> = CyclotomicField(5)
        sage: A = matrix(3, 3, [1,z,z^2,z^3,z^4,2/3*z,-3*z,z,2+z]); B = matrix(3, 3, [-1,2*z,3*z^2,5*z+1,z^4,1/3*z,2-z,3-z,5-z])
        sage: A*B
        [        -z^3 + 7*z^2 + z - 1       -z^3 + 3*z^2 + 2*z + 1              -z^3 + 25/3*z^2]
        [-2*z^3 - 5/3*z^2 + 1/3*z + 4           -z^3 - 8/3*z^2 - 2     -2/3*z^2 + 10/3*z + 10/3]
        [             4*z^2 + 4*z + 4               -7*z^2 + z + 7  -9*z^3 - 2/3*z^2 + 3*z + 10]

    Verify that the answer above is consistent with what the
    generic sparse matrix multiply gives (which is a different
    implementation).::

        sage: A*B == A.sparse_matrix()*B.sparse_matrix()
        True

        sage: N1 = Matrix(CyclotomicField(6), 1, [1])
        sage: cf6 = CyclotomicField(6) ; z6 = cf6.0
        sage: N2 = Matrix(CyclotomicField(6), 1, 5, [0,1,z6,-z6,-z6+1])
        sage: N1*N2
        [         0          1      zeta6     -zeta6 -zeta6 + 1]
        sage: N1 = Matrix(CyclotomicField(6), 1, [-1])
        sage: N1*N2
        [        0        -1    -zeta6     zeta6 zeta6 - 1]

    Verify that a degenerate case bug reported at :issue:`5974` is fixed.

        sage: K.<zeta6>=CyclotomicField(6); matrix(K,1,2) * matrix(K,2,[0, 1, 0, -2*zeta6, 0, 0, 1, -2*zeta6 + 1])
        [0 0 0 0]

    TESTS:

    This is from :issue:`8666`::

        sage: K.<zeta4> = CyclotomicField(4)
        sage: m = matrix(K, [125])
        sage: n = matrix(K, [186])
        sage: m*n
        [23250]
        sage: (-m)*n
        [-23250]
    """
    A, denom_self = self._matrix._clear_denom()
    B, denom_right = (<Matrix_cyclo_dense>right)._matrix._clear_denom()

    # conservative but correct estimate: 2 is there to account for the
    # sign of the entries
    bound = 1 + 2 * A.height() * B.height() * self._ncols

    n = self._base_ring._n()
    p = previous_prime(MAX_MODULUS)
    prod = 1
    v = []
    while prod <= bound:
        while (n >= 2 and p % n != 1) or denom_self % p == 0 or denom_right % p == 0:
            if p == 2:
                raise RuntimeError("we ran out of primes in matrix multiplication.")
            p = previous_prime(p)
        prod *= p
        Amodp, _ = self._reductions(p)
        Bmodp, _ = right._reductions(p)
        _, S = self._reduction_matrix(p)
        X = Amodp[0]._matrix_from_rows_of_matrices([Amodp[i] * Bmodp[i] for i in range(len(Amodp))])
        v.append(S*X)
        p = previous_prime(p)
    M = matrix(ZZ, self._base_ring.degree(), self._nrows*right.ncols())
    _lift_crt(M, v)
    d = denom_self * denom_right
    if d == 1:
        M = M.change_ring(QQ)
    else:
        M = (1/d)*M
    cdef Matrix_cyclo_dense C = Matrix_cyclo_dense.__new__(Matrix_cyclo_dense,
                MatrixSpace(self._base_ring, self._nrows, right.ncols()),
                                                           None, None, None)
    C._matrix = M
    return C


def _charpoly_multimodular(Matrix_cyclo_dense self, var='x', proof=None):
    """
    Compute the characteristic polynomial of ``self`` using a
    multimodular algorithm.

    INPUT:

    - ``proof`` -- boolean (default: global flag); if ``False``, compute
      using primes `p_i` until the lift modulo all primes up to `p_i` is
      the same as the lift modulo all primes up to `p_{i+3}` or the bound
      is reached

    EXAMPLES::

        sage: K.<z> = CyclotomicField(3)
        sage: A = matrix(3, [-z, 2*z + 1, 1/2*z + 2, 1, -1/2, 2*z + 2, -2*z - 2, -2*z - 2, 2*z - 1])
        sage: A._charpoly_multimodular()
        x^3 + (-z + 3/2)*x^2 + (17/2*z + 9/2)*x - 9/2*z - 23/2
        sage: A._charpoly_multimodular('T')
        T^3 + (-z + 3/2)*T^2 + (17/2*z + 9/2)*T - 9/2*z - 23/2
        sage: A._charpoly_multimodular('T', proof=False)
        T^3 + (-z + 3/2)*T^2 + (17/2*z + 9/2)*T - 9/2*z - 23/2

    TESTS:

    We test a degenerate case::

        sage: A = matrix(CyclotomicField(1),2,[1,2,3,4]); A.charpoly()
        x^2 - 5*x - 2
    """
    cdef Matrix_cyclo_dense A
    A = Matrix_cyclo_dense.__new__(Matrix_cyclo_dense, self.parent(),
                                   None, None, None)

    proof = get_proof_flag(proof, "linear_algebra")

    n = self._base_ring._n()
    p = previous_prime(MAX_MODULUS)
    prod = 1
    v = []
    # A, denom = self._matrix._clear_denom()
    # TODO: this might be stupidly slow
    denom = self._matrix.denominator()
    A._matrix = <Matrix_rational_dense>(denom*self._matrix)
    bound = A._charpoly_bound()
    L_last = 0
    while prod <= bound:
        while (n >= 2  and p % n != 1) or denom % p == 0:
            if p == 2:
                raise RuntimeError("we ran out of primes in multimodular charpoly algorithm.")
            p = previous_prime(p)

        X = A._charpoly_mod(p)
        v.append(X)
        prod *= p
        p = previous_prime(p)

        # if we've used enough primes as determined by bound, or
        # if we've used 3 primes, we check to see if the result is
        # the same.
        if prod >= bound or (not proof and (len(v) % 3 == 0)):
            M = matrix(ZZ, self._base_ring.degree(), self._nrows+1)
            L = _lift_crt(M, v)
            if not proof and L == L_last:
                break
            L_last = L

    # Now each column of L encodes a coefficient of the output polynomial,
    # with column 0 being the constant coefficient.
    K = self.base_ring()
    R = K[var]
    coeffs = [K(w.list()) for w in L.columns()]
    f = R(coeffs)

    # Rescale to account for denominator, if necessary
    if denom != 1:
        x = R.gen()
        f = f(x * denom) * (1 / (denom**f.degree()))

    return f


def _echelon_form_multimodular(Matrix_cyclo_dense self, num_primes=10, height_guess=None):
    """
    Use a multimodular algorithm to find the echelon form of ``self``.

    INPUT:

    - ``num_primes`` -- number of primes to work modulo

    - ``height_guess`` -- guess for the height of the echelon form of self

    OUTPUT: matrix in reduced row echelon form

    EXAMPLES::

        sage: W.<z> = CyclotomicField(3)
        sage: A = matrix(W, 2, 3, [1+z, 2/3, 9*z+7, -3 + 4*z, z, -7*z]); A
        [  z + 1     2/3 9*z + 7]
        [4*z - 3       z    -7*z]
        sage: A._echelon_form_multimodular(10)
        [                  1                   0  -192/97*z - 361/97]
        [                  0                   1 1851/97*z + 1272/97]

    TESTS:

    We test a degenerate case::

        sage: A = matrix(CyclotomicField(5),0); A
        []
        sage: A._echelon_form_multimodular(10)
        []
        sage: A.pivots()
        ()

        sage: A = matrix(CyclotomicField(13), 2, 3, [5, 1, 2, 46307, 46307*4, 46307])
        sage: A._echelon_form_multimodular()
        [   1    0 7/19]
        [   0    1 3/19]
    """
    cdef Matrix_cyclo_dense res
    cdef bint is_square

    verbose("entering _echelon_form_multimodular",
            level=echelon_verbose_level)

    denom = self._matrix.denominator()
    A = denom * self

    # This bound is chosen somewhat arbitrarily. Changing it affects the
    # runtime, not the correctness of the result.
    if height_guess is None:
        height_guess = (A.coefficient_bound() + 100) * 1000000

    # This is all setup to keep track of various data
    # in the loop below.
    p = previous_prime(MAX_MODULUS)
    found = 0
    prod = 1
    n = self._base_ring._n()
    height_bound = self._ncols * height_guess * A.coefficient_bound() + 1
    mod_p_ech_ls = []
    max_pivots = tuple()
    is_square = self._nrows == self._ncols

    verbose("using height bound %s" % height_bound,
            level=echelon_verbose_level)

    while True:
        # Generate primes to use, and find echelon form
        # modulo those primes.
        while found < num_primes or prod <= height_bound:
            if (n == 1) or p % n == 1:
                try:
                    mod_p_ech, piv_ls = A._echelon_form_one_prime(p)
                except ValueError:
                    # This means that we chose a prime which divides
                    # the denominator of the echelon form of self, so
                    # just skip it and continue
                    p = previous_prime(p)
                    continue
                # if we have the identity, just return it, and
                # we're done.
                if is_square and len(piv_ls) == self._nrows:
                    self.cache('pivots', tuple(range(self._nrows)))
                    return self.parent().identity_matrix()
                if piv_ls > max_pivots:
                    mod_p_ech_ls = [mod_p_ech]
                    max_pivots = piv_ls
                    # add this to the list of primes
                    prod = p
                    found = 1
                elif piv_ls == max_pivots:
                    mod_p_ech_ls.append(mod_p_ech)
                    # add this to the list of primes
                    prod *= p
                    found += 1
                else:
                    # this means that the rank profile mod this
                    # prime is worse than those that came before,
                    # so we just loop
                    p = previous_prime(p)
                    continue

            p = previous_prime(p)

        if found > num_primes:
            num_primes = found

        verbose("computed echelon form mod %s primes" % num_primes,
                level=echelon_verbose_level)
        verbose("current product of primes used: %s" % prod,
                level=echelon_verbose_level)

        # Use CRT to lift back to ZZ
        mat_over_ZZ = matrix(ZZ, self._base_ring.degree(),
                             self._nrows * self._ncols)
        _lift_crt(mat_over_ZZ, mod_p_ech_ls)
        # note: saving the CRT intermediate MultiModularBasis does
        # not seem to affect the runtime at all

        # Attempt to use rational reconstruction to find
        # our echelon form
        try:
            verbose("attempting rational reconstruction ...",
                    level=echelon_verbose_level)
            res = Matrix_cyclo_dense.__new__(Matrix_cyclo_dense,
                                             self.parent(),
                                             None, None, None)
            res._matrix = <Matrix_rational_dense>matrix_integer_dense_rational_reconstruction(mat_over_ZZ, prod)

        except ValueError:
            # If a ValueError is raised here, it means that the
            # rational reconstruction failed. In this case, add
            # on a few more primes, and try again.

            num_primes += echelon_primes_increment
            verbose("rational reconstruction failed, trying with %s primes"%num_primes, level=echelon_verbose_level)
            continue

        verbose("rational reconstruction succeeded with %s primes!"%num_primes, level=echelon_verbose_level)

        if ((res * res.denominator()).coefficient_bound() *
            self.coefficient_bound() * self.ncols()) > prod:
            # In this case, we don't know the result to sufficient
            # "precision" (here precision is just the modulus,
            # prod) to guarantee its correctness, so loop.

            num_primes += echelon_primes_increment
            verbose("height not sufficient to determine echelon form",
                    level=echelon_verbose_level)
            continue

        verbose("found echelon form with %s primes, whose product is %s"%(num_primes, prod), level=echelon_verbose_level)
        self.cache('pivots', max_pivots)
        return res



# sage_setup: distribution = sagemath-linbox
r"""
Inline conversions between LinBox and Sage

Each LinBox type has a corresponding Sage types and we use the following
conventions for conversion functions

- ``new_linbox_XXX`` : create a new linbox object
- ``new_sage_XXX``   : create a new Sage object
- ``set_linbox_XXX`` : set the entries of the linbox object
- ``set_sage_XXX``   : set the entries of the Sage object

For matrices that uses a flint datastructure, see the lower level conversions
in the module ``linbox_flint_interface``.
"""
#*****************************************************************************
#       Copyright (C) 2007 Martin Albrecht
#       Copyright (C) 2008 Clement Pernet
#       Copyright (C) 2018 Vincent Delecroix
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  http://www.gnu.org/licenses/
#*****************************************************************************


from libcpp.vector cimport vector as cppvector
from libcpp.utility cimport move

from sage.libs.gmp.mpz cimport mpz_set
from sage.libs.gmp.mpq cimport mpq_get_num, mpq_get_den

from sage.libs.linbox.givaro cimport Modular_uint64, ZRing, Integer, QField, Rational
from sage.libs.linbox.linbox cimport SparseMatrix_Modular_uint64, SparseMatrix_integer, SparseMatrix_rational, DenseVector_integer

from sage.matrix.matrix_modn_sparse cimport Matrix_modn_sparse
from sage.matrix.matrix_integer_sparse cimport Matrix_integer_sparse
from sage.matrix.matrix_rational_sparse cimport Matrix_rational_sparse

from sage.modules.vector_modn_sparse cimport c_vector_modint
from sage.modules.vector_integer_dense cimport Vector_integer_dense
from sage.modules.vector_integer_sparse cimport mpz_vector
from sage.modules.vector_rational_sparse cimport mpq_vector

########################################
# algorithm for solving linear systems #
########################################

ctypedef enum linbox_specifier:
    METHOD_DEFAULT              # no specification
    METHOD_DENSE_ELIMINATION     # DenseElimination
    METHOD_SPARSE_ELIMINATION   # SparseElimination
    METHOD_BLACKBOX             # Blackbox
    METHOD_WIEDEMANN            # Wiedeman
    ERROR

cdef inline linbox_specifier get_method(str algo) except ERROR:
    if algo is None or algo == "default":
        return METHOD_DEFAULT
    elif algo == "dense_elimination" or \
         algo == "linbox_dense_elimination" or \
         algo == "LinBox::DenseElimination":
        return METHOD_DENSE_ELIMINATION
    elif algo == "sparse_elimination" or \
         algo == "linbox_sparse_elimination" or \
         algo == "LinBox::SparseElimination":
        return METHOD_SPARSE_ELIMINATION
    elif algo == "blackbox" or \
         algo == "linbox_blackbox" or \
         algo == "LinBox::Blackbox":
        return METHOD_BLACKBOX
    elif algo == 'wiedemann' or \
         algo == "linbox_wiedemann" or \
         algo == "LinBox::Wiedeman":
        return METHOD_WIEDEMANN
    else:
        raise ValueError("unknown algorithm")

################################################
# matrix_modn_sparse (sparse matrix over Z/nZ) #
################################################

cdef inline void set_linbox_matrix_modn_sparse(SparseMatrix_Modular_uint64& A, Matrix_modn_sparse m) noexcept:
    r"""
    Set the entries of a LinBox matrix from a Sage matrix.

    INPUT:

    - ``A`` -- LinBox matrix
    - ``m`` -- Sage matrix
    """
    cdef c_vector_modint * row
    cdef size_t i, j
    for i in range(<size_t> m._nrows):
        row = m.rows + i
        for j in range(<size_t> row.num_nonzero):
            A.setEntry(i, row.positions[j], row.entries[j])

cdef inline SparseMatrix_Modular_uint64 * new_linbox_matrix_modn_sparse(Modular_uint64 &F, Matrix_modn_sparse m) noexcept:
    r"""
    Return a new LinBox matrix from a Sage matrix.

    Such matrix has to be deallocated with a "del" statement.

    INPUT:

    - ``F`` -- LinBox field
    - ``m`` -- Sage matrix
    """
    cdef SparseMatrix_Modular_uint64 * A = new SparseMatrix_Modular_uint64(F, <size_t> m._nrows, <size_t> m._ncols)
    set_linbox_matrix_modn_sparse(A[0], m)
    return A

#########################
# matrix_integer_sparse #
#########################

cdef inline void set_linbox_matrix_integer_sparse(SparseMatrix_integer& A, Matrix_integer_sparse m) noexcept:
    r"""
    Set the entries of a LinBox matrix from a Sage matrix.

    INPUT:

    - ``A`` -- LinBox matrix
    - ``m`` -- Sage matrix
    """
    cdef size_t i, j, k
    cdef mpz_vector * v
    cdef Integer t
    for i in range(<size_t> m._nrows):
        v = m._matrix + i
        for k in range(<size_t> v.num_nonzero):
            j = v.positions[k]
            mpz_set(t.get_mpz(), v.entries[k])
            A.setEntry(i, j, t)

cdef inline SparseMatrix_integer * new_linbox_matrix_integer_sparse(ZRing &ZZ, Matrix_integer_sparse m) noexcept:
    r"""
    Return a new LinBox matrix from a Sage matrix.

    Such a matrix has to be deallocated with a "del" statement.

    INPUT:

    - ``m`` -- Sage matrix
    """
    cdef SparseMatrix_integer * A = new SparseMatrix_integer(ZZ, <size_t> m._nrows, <size_t> m._ncols)
    set_linbox_matrix_integer_sparse(A[0], m)
    return A

##########################
# matrix rational sparse #
##########################

cdef inline void set_linbox_matrix_rational_sparse(SparseMatrix_rational& A, Matrix_rational_sparse m) noexcept:
    r"""
    Set the entries of a LinBox matrix from a Sage matrix.

    INPUT:

    - ``A`` -- LinBox matrix
    - ``m`` -- Sage matrix
    """
    cdef size_t i, j, k
    cdef mpq_vector * v
    cdef Integer num, denom
    for i in range(<size_t> m._nrows):
        v = m._matrix + i
        for k in range(<size_t> v.num_nonzero):
            j = v.positions[k]
            mpq_get_num(num.get_mpz(), v.entries[k])
            mpq_get_den(denom.get_mpz(), v.entries[k])
            A.setEntry(i, j, Rational(move(num), move(denom), 0))

cdef inline SparseMatrix_rational * new_linbox_matrix_rational_sparse(QField &QQ, Matrix_rational_sparse m) noexcept:
    r"""
    Return a new LinBox matrix from a Sage matrix.

    Such a matrix has to be deallocated with a "del" statement.

    INPUT:

    - ``m`` -- Sage matrix
    """
    cdef SparseMatrix_rational * A = new SparseMatrix_rational(QQ, <size_t> m._nrows, <size_t> m._ncols)
    set_linbox_matrix_rational_sparse(A[0], m)
    return A

########################
# vector integer dense #
########################

cdef inline DenseVector_integer * new_linbox_vector_integer_dense(ZRing &ZZ, Vector_integer_dense v) noexcept:
    r"""
    Return a new linbox vector from a sage one.

    INPUT:

    - ``v`` -- a Sage dense integer vector
    """
    cdef cppvector[Integer] * vec = new cppvector[Integer](<size_t> v._degree)
    cdef size_t i
    for i in range(<size_t> v._degree):
        mpz_set(vec[0][i].get_mpz(), v._entries[i])

    cdef DenseVector_integer * V = new DenseVector_integer(ZZ, vec[0])
    del vec

    return V

cdef inline Vector_integer_dense new_sage_vector_integer_dense(P, DenseVector_integer &v):
    r"""
    Return a new Sage vector from a LinBox one.

    INPUT:

    - ``P`` -- parent for the Sage vector
    - ``v`` -- linbox vector
    """
    cdef Vector_integer_dense res = P()
    cdef size_t i
    for i in range(<size_t> res._degree):
        mpz_set(res._entries[i], v.getEntry(i).get_mpz_const())

    return res



# sage_setup: distribution = sagemath-linbox
# distutils: extra_compile_args = LINBOX_CFLAGS -std=gnu++14
# distutils: include_dirs = LINBOX_INCDIR
# distutils: libraries = LINBOX_LIBRARIES
# distutils: library_dirs = LINBOX_LIBDIR
# distutils: extra_link_args = LINBOX_LIBEXTRA
# distutils: language = c++

from libc.stdint cimport uint32_t, uint64_t
from libcpp.vector cimport vector as cppvector
from libcpp.pair cimport pair

from sage.libs.linbox.givaro cimport *

cdef extern from "linbox/matrix/dense-matrix.h":
    ## template <class _Field, class _blasRep=typename Vector<_Field>::Dense >
    ## class DenseMatrix ;
    ##
    ## template <class _Field>
    ## using DenseMatrix = DenseMatrix<_Field> ;
    cdef cppclass DenseMatrix_integer "LinBox::DenseMatrix<Givaro::ZRing<Givaro::Integer>>":
        ctypedef ZRing Field
        ctypedef Integer Element
        DenseMatrix_integer(Field &F, size_t m, size_t n)
        size_t rowdim()
        size_t coldim()
        void setEntry(size_t i, size_t j, Element &a)
        Element &getEntry(size_t i, size_t j)
        Field& field()

        ostream& write(ostream&)

    cdef cppclass DenseMatrix_Modular_double "LinBox::DenseMatrix<Givaro::Modular<double>>":
        ctypedef Modular_double Field
        ctypedef double Element
        DenseMatrix_Modular_double(Field F, size_t m, size_t n)
        void setEntry(size_t i, size_t j, Element& a)
        Element &getEntry(size_t i, size_t j)

        ostream& write(ostream&)

    cdef cppclass DenseMatrix_Modular_float "LinBox::DenseMatrix<Givaro::Modular<float>>":
        ctypedef Modular_float Field
        ctypedef float Element
        DenseMatrix_Modular_float(Field F, size_t m, size_t n)
        void setEntry(size_t i, size_t j, Element& a)
        Element &getEntry(size_t i, size_t j)

        ostream& write(ostream&)

cdef extern from "linbox/vector/sparse.h":
    cdef cppclass Sparse_Vector_rational "LinBox::Sparse_Vector<Givaro::Rational>":
        ctypedef pair[unsigned int, Rational] Element
        size_t size()
        Element& operator[](size_t i)

cdef extern from "linbox/matrix/sparse-matrix.h":
    ## template<class _Field, class _Storage = SparseMatrixFormat::SparseSeq >
    ## class SparseMatrix ;
    cdef cppclass SparseMatrix_integer "LinBox::SparseMatrix<Givaro::ZRing<Givaro::Integer>>":
        ctypedef ZRing Field
        ctypedef Integer Element
        SparseMatrix_integer(Field &F, size_t m, size_t n)
        size_t rowdim()
        size_t coldim()
        void setEntry(size_t i, size_t j, Element &a)
        Element &getEntry(size_t i, size_t j)
        Field& field()

        ostream& write(ostream&)

    cdef cppclass SparseMatrix_Modular_uint64 "LinBox::SparseMatrix<Givaro::Modular<uint64_t>, LinBox::SparseMatrixFormat::SparseSeq>":
        ctypedef Modular_uint64 Field
        ctypedef uint64_t Element
        SparseMatrix_Modular_uint64(Field &F, size_t m, size_t n)
        SparseMatrix_Modular_uint64(SparseMatrix_Modular_uint64&)
        size_t rowdim()
        size_t coldim()
        void setEntry(size_t i, size_t j, Element &a)
        Element &getEntry(size_t i, size_t j)
        Field& field()

        ostream& write(ostream&)

    cdef cppclass SparseMatrix_rational "LinBox::SparseMatrix<Givaro::QField<Givaro::Rational>>":
        ctypedef QField Field
        ctypedef Rational Element
        SparseMatrix_rational(Field &F, size_t m, size_t n)
        size_t rowdim()
        size_t coldim()
        void setEntry(size_t i, size_t j, Element &a)
        Element &getEntry(size_t i, size_t j)
        Field& field()
        Sparse_Vector_rational& getRow(size_t i)

        ostream& write(ostream&)

cdef extern from "linbox/polynomial/dense-polynomial.h":
    ## template<class Field>
    ## class DensePolynomial : public Givaro::Poly1FactorDom<Field, Givaro::Dense>::Element
    cdef cppclass DensePolynomial_integer "LinBox::DensePolynomial<Givaro::ZRing<Givaro::Integer>>":
        ctypedef ZRing BaseRing
        ctypedef Integer BaseRingElement
        DensePolynomial_integer(BaseRing &F)
        DensePolynomial_integer(BaseRing &F, size_t s)
        BaseRingElement& operator[](size_t i)
        size_t size()

cdef extern from "linbox/ring/polynomial-ring.h":
    ## template <class BaseRing, class StorageTag= Givaro::Dense>
    ## class PolynomialRing : public Givaro::Poly1FactorDom<BaseRing,StorageTag>
    cdef cppclass PolynomialRing_integer "LinBox::PolynomialRing<Givaro::ZRing<Givaro::Integer>, Givaro::Dense>":
        ctypedef DensePolynomial_integer Element
        ctypedef DensePolynomial_integer Polynomial

cdef extern from "linbox/vector/vector.h":
    cdef cppclass DenseVector_integer "LinBox::DenseVector<Givaro::ZRing<Givaro::Integer>>":
        ctypedef ZRing Field
        ctypedef Integer Element
        DenseVector_integer (Field &F)
        DenseVector_integer (Field &F, long& m)
        DenseVector_integer (Field &F, cppvector[Integer]&)
        size_t size()
        void resize(size_t)
        void resize(size_t n, const Element&)
        void push_back(Element&)
        void clear()
        void reserve(const size_t&)
        void setEntry(size_t i, Element&)
        Element& getEntry(size_t i)

cdef extern from "linbox/matrix/matrix-domain.h":
    ## template <class Field_ >
    ## class MatrixDomain : public MVProductDomain<Field_> {
    cdef cppclass MatrixDomain_integer "LinBox::MatrixDomain<Givaro::ZRing<Givaro::Integer>>":
        MatrixDomain_integer(ZRing&)
        DenseMatrix_integer& mul(DenseMatrix_integer ans,
                                      DenseMatrix_integer left,
                                      DenseMatrix_integer right)

cdef extern from "linbox/solutions/methods.h" namespace "LinBox":
    cdef struct HybridSpecifier:
        pass
    cdef struct BlackboxSpecifier:
        pass
    cdef struct EliminationSpecifier:
        pass
    cdef struct WiedemannTraits:
        pass
    cdef struct DenseEliminationTraits:
        pass
    cdef struct SparseEliminationTraits:
        pass

    cdef cppclass Method:
        ctypedef HybridSpecifier Hybrid
        ctypedef BlackboxSpecifier Blackbox
        ctypedef EliminationSpecifier Elimination
        ctypedef WiedemannTraits Wiedemann
        ctypedef DenseEliminationTraits DenseElimination
        ctypedef SparseEliminationTraits SparseElimination

cdef extern from "linbox/solutions/charpoly.h" namespace "LinBox":
    PolynomialRing_integer.Element& charpoly (PolynomialRing_integer.Element&, DenseMatrix_integer&)
    PolynomialRing_integer.Element& charpoly (PolynomialRing_integer.Element&, SparseMatrix_integer&)

cdef extern from "linbox/solutions/minpoly.h" namespace "LinBox":
    PolynomialRing_integer.Element& minpoly (PolynomialRing_integer.Element&, DenseMatrix_integer&)
    PolynomialRing_integer.Element& minpoly (PolynomialRing_integer.Element&, SparseMatrix_integer&)

cdef extern from "linbox/algorithms/gauss.h":
    cdef cppclass GaussDomain_Modular_uint64 "LinBox::GaussDomain<Givaro::Modular<uint64_t>>":
        ctypedef Modular_uint64 Field
        ctypedef uint64_t Element
        GaussDomain_Modular_uint64(Field &)
        unsigned long& InPlaceLinearPivoting(unsigned long &rank,
                                             Element& determinant,
                                             SparseMatrix_Modular_uint64 &A,
                                             unsigned long Ni,
                                             unsigned long Nj)

    cdef cppclass GaussDomain_rational "LinBox::GaussDomain<Givaro::QField<Givaro::Rational>>":
        ctypedef QField Field
        ctypedef Rational Element
        GaussDomain_rational(Field &)
        SparseMatrix_rational& nullspacebasisin(SparseMatrix_rational &X, SparseMatrix_rational &A)

cdef extern from "linbox/solutions/echelon.h" namespace "LinBox":
    size_t rowEchelon (DenseMatrix_Modular_float&, const DenseMatrix_Modular_float&)
    size_t rowEchelonize (DenseMatrix_Modular_float&)
    size_t reducedRowEchelon (DenseMatrix_Modular_float&, const DenseMatrix_Modular_float&)
    size_t reducedRowEchelonize (DenseMatrix_Modular_float&)
    size_t colEchelon (DenseMatrix_Modular_float&, const DenseMatrix_Modular_float&)
    size_t colEchelonize (DenseMatrix_Modular_float&)
    size_t reducedColEchelon (DenseMatrix_Modular_float&, const DenseMatrix_Modular_float&)
    size_t reducedColEchelonize (DenseMatrix_Modular_float&)

    size_t rowEchelon (DenseMatrix_Modular_double&, const DenseMatrix_Modular_double&)
    size_t rowEchelonize (DenseMatrix_Modular_double&)
    size_t reducedRowEchelon (DenseMatrix_Modular_double&, const DenseMatrix_Modular_double&)
    size_t reducedRowEchelonize (DenseMatrix_Modular_double&)
    size_t colEchelon (DenseMatrix_Modular_double&, const DenseMatrix_Modular_double&)
    size_t colEchelonize (DenseMatrix_Modular_double&)
    size_t reducedColEchelon (DenseMatrix_Modular_double&, const DenseMatrix_Modular_double&)
    size_t reducedColEchelonize (DenseMatrix_Modular_double&)

cdef extern from "linbox/solutions/rank.h" namespace "LinBox":
    unsigned long & rank (unsigned long&, DenseMatrix_integer)
    unsigned long & rank (unsigned long&, SparseMatrix_integer)

cdef extern from "linbox/solutions/det.h" namespace "LinBox":
    Integer& det (Integer&, DenseMatrix_integer)
    Integer& det (Integer&, SparseMatrix_integer)

cdef extern from "linbox/solutions/solve.h" namespace "LinBox":
    # integer solve

    DenseVector_integer& solve (DenseVector_integer &,
                                Integer &,
                                SparseMatrix_integer &,
                                DenseVector_integer &,
                                Method.DenseElimination) except +

    DenseVector_integer& solve (DenseVector_integer &,
                                Integer &,
                                SparseMatrix_integer &,
                                DenseVector_integer &,
                                Method.SparseElimination) except +

    DenseVector_integer& solve (DenseVector_integer &,
                                Integer &,
                                SparseMatrix_integer &,
                                DenseVector_integer &,
                                Method.Blackbox) except +

    DenseVector_integer& solve (DenseVector_integer &,
                                Integer &,
                                SparseMatrix_integer &,
                                DenseVector_integer &,
                                Method.Wiedemann) except +

    DenseVector_integer& solve (DenseVector_integer &,
                                Integer &,
                                SparseMatrix_integer &,
                                DenseVector_integer &)



# sage_setup: distribution = sagemath-linbox

from cysignals.signals cimport sig_on, sig_off

cimport sage.libs.linbox.givaro as givaro
cimport sage.libs.linbox.linbox as linbox
import sage.matrix.matrix_space

from sage.libs.gmp.mpq cimport *
from sage.libs.linbox.conversion cimport new_linbox_matrix_rational_sparse
from sage.matrix.matrix_rational_sparse cimport Matrix_rational_sparse
from sage.modules.vector_rational_sparse cimport *
from sage.rings.rational_field import QQ


def _right_kernel_matrix_linbox(Matrix_rational_sparse self):
    r"""
    Return a pair that includes a matrix of basis vectors
    for the right kernel of ``self``.

    OUTPUT:

    Returns a pair.  First item is the string 'computed-linbox-rational'
    that identifies the nature of the basis vectors.

    Second item is a matrix whose rows are a basis for the right kernel,
    over the rationals, as computed by the LinBox library.
    """
    # NOTE: Degenerate cases (0 rows or columns) are handled by right_kernel_matrix.

    cdef givaro.QField givQQ
    cdef linbox.SparseMatrix_rational * M = new_linbox_matrix_rational_sparse(givQQ, self)

    MQ = sage.matrix.matrix_space.MatrixSpace(QQ, self._ncols, self._ncols, sparse=True)
    A = MQ.zero_matrix().__copy__()

    cdef linbox.SparseMatrix_rational * N = new_linbox_matrix_rational_sparse(givQQ, A)

    cdef linbox.GaussDomain_rational * dom = new linbox.GaussDomain_rational(givQQ)

    cdef Matrix_rational_sparse ans
    cdef mpq_t s

    try:
        sig_on()
        dom.nullspacebasisin(N[0], M[0])
        sig_off()
    except KeyboardInterrupt:
        del N
        raise
    finally:
        del M, dom

    ans = self.new_matrix(N.coldim(), N.rowdim()) # NOTE: Transposing.
    cdef size_t i, k
    cdef Py_ssize_t j
    mpq_init(s)
    for i in range(N.rowdim()):
        for k in range(N.getRow(i).size()):
            j = <Py_ssize_t> N.getRow(i)[k].first
            entry = N.getRow(i)[k].second
            mpq_set_num(s, entry.nume().get_mpz())
            mpq_set_den(s, entry.deno().get_mpz())
            mpq_vector_set_entry(&ans._matrix[j], i, s) # NOTE: Transposing.
    mpq_clear(s)

    del N

    # Remove zero rows, if any.
    if ans.nrows() > 0:
        r = ans.nrows() - 1
        while r >= 0 and ans.row(r) == 0:
            r -= 1
        ans = ans.submatrix(0, 0, r + 1)

    return 'computed-linbox-rational', ans



# sage_setup: distribution = sagemath-linbox
# distutils: libraries = LINBOX_LIBRARIES
# distutils: library_dirs = LINBOX_LIBDIR
# distutils: extra_link_args = LINBOX_LIBEXTRA
r"""
Interface between flint matrices and linbox

This module only contains C++ code (and the interface is fully C
compatible). It basically contains what used to be in the LinBox
source code under interfaces/sage/linbox-sage.C written by M. Albrecht
and C. Pernet. The functions available are:

- ``void  linbox_fmpz_mat_mul(fmpz_mat_t C, fmpz_mat_t A, fmpz_mat_t B)``: set
  ``C`` to be the result of the multiplication ``A * B``

- ``void linbox_fmpz_mat_charpoly(fmpz_poly_t cp, fmpz_mat_t A)``: set ``cp``
  to be the characteristic polynomial of the square matrix ``A``

- ``void  linbox_fmpz_mat_minpoly(fmpz_poly_t mp, fmpz_mat_t A)``: set ``mp``
  to be the minimal polynomial of the square matrix ``A``

- ``size_t linbox_fmpz_mat_rank(fmpz_mat_t A)``: return the rank of the
  matrix ``A``

- ``void linbox_fmpz_mat_det(fmpz_t det, fmpz_mat_t A)``: set ``det`` to the
  determinant of the square matrix ``A``
"""
# ****************************************************************************
#       Copyright (C) 2007 Martin Albrecht
#       Copyright (C) 2008 Clement Pernet
#       Copyright (C) 2017-2018 Vincent Delecroix
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.libs.gmp.types cimport mpz_t
from sage.libs.flint.types cimport fmpz_t
from sage.libs.flint.fmpz cimport fmpz_get_mpz, fmpz_set_mpz
from sage.libs.flint.fmpz_mat cimport fmpz_mat_entry, fmpz_mat_nrows, fmpz_mat_ncols
from sage.libs.flint.fmpz_poly cimport fmpz_poly_fit_length, _fmpz_poly_set_length
from sage.libs.flint.fmpz_poly_sage cimport fmpz_poly_set_coeff_mpz

cimport sage.libs.linbox.givaro as givaro
cimport sage.libs.linbox.linbox as linbox
from sage.libs.linbox.linbox cimport PolynomialRing_integer


cdef void fmpz_mat_get_linbox(linbox.DenseMatrix_integer& A, fmpz_mat_t m) noexcept:
    r"""
    Set the entries of A from m (no allocation performed).

    NOTE: this function does not appear in the Cython header
    (the .pxd file) in order to keep the header C-compatible
    """
    cdef size_t i,j
    cdef givaro.Integer t

    for i in range(fmpz_mat_nrows(m)):
        for j in range(fmpz_mat_ncols(m)):
            fmpz_get_mpz(t.get_mpz(), fmpz_mat_entry(m, i, j))
            A.setEntry(i, j, t)


cdef void fmpz_mat_set_linbox(fmpz_mat_t m, linbox.DenseMatrix_integer& A) noexcept:
    r"""
    Set the entries of m from A (no allocation performed).

    NOTE: this function does not appear in the Cython header
    (the .pxd file) in order to keep the header C-compatible
    """
    cdef size_t i,j
    for i in range(A.rowdim()):
        for j in range(A.coldim()):
            fmpz_set_mpz(fmpz_mat_entry(m, i, j), A.getEntry(i, j).get_mpz_const())


cdef void fmpz_poly_set_linbox(fmpz_poly_t p, PolynomialRing_integer.Element& q) noexcept:
    r"""
    Set the entries of the polynomial p from q (no allocation performed).

    NOTE: this function does not appear in the Cython header
    (the .pxd file) in order to keep the header C-compatible
    """
    cdef size_t i
    cdef mpz_t tmp

    fmpz_poly_fit_length(p, q.size())

    for i in range(q.size()):
        tmp = q[i].get_mpz_const()
        fmpz_poly_set_coeff_mpz(p, i, tmp)

    _fmpz_poly_set_length(p, q.size())


cdef void linbox_fmpz_mat_mul(fmpz_mat_t C, fmpz_mat_t A, fmpz_mat_t B) noexcept:
    r"""
    Set C to be A * B.
    """
    cdef givaro.ZRing ZZ
    cdef linbox.DenseMatrix_integer *LBA
    cdef linbox.DenseMatrix_integer *LBB
    cdef linbox.DenseMatrix_integer *LBC
    cdef linbox.MatrixDomain_integer * MD

    LBA = new linbox.DenseMatrix_integer(ZZ, fmpz_mat_nrows(A), fmpz_mat_ncols(A))
    fmpz_mat_get_linbox(LBA[0], A)

    LBB = new linbox.DenseMatrix_integer(ZZ, fmpz_mat_nrows(B), fmpz_mat_ncols(B))
    fmpz_mat_get_linbox(LBB[0], B)

    LBC = new linbox.DenseMatrix_integer(ZZ, fmpz_mat_nrows(A), fmpz_mat_ncols(B))

    MD = new linbox.MatrixDomain_integer(ZZ)
    MD.mul(LBC[0], LBA[0], LBB[0])

    del MD

    fmpz_mat_set_linbox(C, LBC[0])


cdef void linbox_fmpz_mat_charpoly(fmpz_poly_t cp, fmpz_mat_t A) noexcept:
    r"""
    Set cp to the characteristic polynomial of A.
    """
    cdef givaro.ZRing ZZ
    cdef linbox.DenseMatrix_integer * LBA
    cdef linbox.DensePolynomial_integer * m_A

    LBA = new linbox.DenseMatrix_integer(ZZ, fmpz_mat_nrows(A), fmpz_mat_ncols(A))
    fmpz_mat_get_linbox(LBA[0], A)
    m_A = new linbox.DensePolynomial_integer(ZZ, fmpz_mat_nrows(A))
    linbox.charpoly(m_A[0], LBA[0])
    fmpz_poly_set_linbox(cp, m_A[0])

    del LBA
    del m_A


cdef void linbox_fmpz_mat_minpoly(fmpz_poly_t mp, fmpz_mat_t A) noexcept:
    r"""
    Set mp to the minimal polynomial of A.
    """
    cdef givaro.ZRing ZZ
    cdef linbox.DenseMatrix_integer * LBA
    cdef linbox.DensePolynomial_integer * m_A

    LBA = new linbox.DenseMatrix_integer(ZZ, fmpz_mat_nrows(A), fmpz_mat_ncols(A))
    m_A = new linbox.DensePolynomial_integer(ZZ)
    fmpz_mat_get_linbox(LBA[0], A)
    linbox.minpoly(m_A[0], LBA[0])
    fmpz_poly_set_linbox(mp, m_A[0])

    del LBA
    del m_A


cdef size_t linbox_fmpz_mat_rank(fmpz_mat_t A) noexcept:
    r"""
    Return the rank of A
    """
    cdef givaro.ZRing ZZ
    cdef linbox.DenseMatrix_integer * LBA
    cdef size_t r = 0

    LBA = new linbox.DenseMatrix_integer(ZZ, fmpz_mat_nrows(A), fmpz_mat_ncols(A))
    fmpz_mat_get_linbox(LBA[0], A)
    linbox.rank(r, LBA[0])

    del LBA

    return r


cdef void linbox_fmpz_mat_det(fmpz_t det, fmpz_mat_t A) noexcept:
    r"""
    Set det to the determinant of A.
    """
    cdef givaro.ZRing ZZ
    cdef linbox.DenseMatrix_integer * LBA
    cdef givaro.Integer d

    LBA = new linbox.DenseMatrix_integer(ZZ, fmpz_mat_nrows(A), fmpz_mat_ncols(A))
    fmpz_mat_get_linbox(LBA[0], A)
    linbox.det(d, LBA[0])
    fmpz_set_mpz(det, d.get_mpz_const())

    del LBA



# sage_setup: distribution = sagemath-ntl
from sage.libs.ntl.types cimport GF2EContext_c
from sage.libs.ntl.ntl_GF2X cimport ntl_GF2X

cdef class ntl_GF2EContext_class():
    cdef GF2EContext_c x
    cdef ntl_GF2X m
    cdef void restore_c(self) noexcept
    cdef object __weakref__



# sage_setup: distribution = sagemath-ntl
from sage.libs.ntl.types cimport GF2E_c, GF2X_c, GF2_c, GF2XModulus_c, ZZ_c


cdef extern from "ntlwrap.h":
    void GF2E_init "GF2E::init"(GF2X_c x)
    long GF2E_degree "GF2E::degree"()
    GF2XModulus_c GF2E_modulus "GF2E::modulus"()

    int GF2E_IsOne "IsOne"(GF2E_c x)
    int GF2E_IsZero "IsZero"(GF2E_c x)

    void GF2E_add "add"( GF2E_c x, GF2E_c a, GF2E_c b)
    void GF2E_sub "sub"( GF2E_c x, GF2E_c a, GF2E_c b)
    void GF2E_mul "mul"( GF2E_c x, GF2E_c a, GF2E_c b)
    void GF2E_div "div"( GF2E_c x, GF2E_c a, GF2E_c b)
    void GF2E_power "NTL::power"(GF2E_c t, GF2E_c x, long e)
    long GF2E_deg "deg"(GF2E_c x)

    void GF2E_conv_GF2X "conv" (GF2E_c out, GF2X_c inp)
    void GF2E_conv_long "conv" (GF2E_c out, long inp)
    void GF2E_conv_ZZ "conv" (GF2E_c out, ZZ_c inp)
    void GF2E_conv_GF2 "conv" (GF2E_c out, GF2_c inp)
    GF2X_c GF2E_rep "rep"(GF2E_c x)

    GF2E_c GF2E_random "random_GF2E"()

    GF2_c GF2E_trace "trace"(GF2E_c x)



# sage_setup: distribution = sagemath-ntl
# distutils: depends = NTL/ZZ.h

from sage.libs.ntl.types cimport ZZ_c, ZZ_p_c, ZZ_pX_c, ZZ_pE_c


cdef extern from "ntlwrap.h":
    void ZZ_pE_add "add"( ZZ_pE_c x, ZZ_pE_c a, ZZ_pE_c b)
    void ZZ_pE_add_long "add"( ZZ_pE_c x, ZZ_pE_c a, long b)
    void ZZ_pE_add_ZZ_p "add"( ZZ_pE_c x, ZZ_pE_c a, ZZ_p_c b)
    void ZZ_pE_sub "sub"( ZZ_pE_c x, ZZ_pE_c a, ZZ_pE_c b)
    void ZZ_pE_sub_long "add"( ZZ_pE_c x, ZZ_pE_c a, long b)
    void ZZ_pE_sub_ZZ_p "sub"( ZZ_pE_c x, ZZ_pE_c a, ZZ_p_c b)
    void ZZ_pE_mul "mul"( ZZ_pE_c x, ZZ_pE_c a, ZZ_pE_c b)
    void ZZ_pE_mul_long "mul"( ZZ_pE_c x, ZZ_pE_c a, long b)
    void ZZ_pE_mul_ZZ_p "mul"( ZZ_pE_c x, ZZ_pE_c a, ZZ_p_c b)
    void ZZ_pE_negate "NTL::negate"(ZZ_pE_c x, ZZ_pE_c a)
    void ZZ_pE_power "NTL::power"(ZZ_pE_c t, ZZ_pE_c x, long e)
    int ZZ_pE_IsOne "IsOne"(ZZ_pE_c x)
    int ZZ_pE_IsZero "IsZero"(ZZ_pE_c x)
    ZZ_pX_c ZZ_pE_rep "rep"(ZZ_pE_c z)
    void ZZ_pE_inv "inv"( ZZ_pE_c x, ZZ_pE_c a) # raises an error if a not invertible
    void ZZ_pE_div "div"( ZZ_pE_c x, ZZ_pE_c a, ZZ_pE_c b) # raises an error if b not invertible
    void ZZ_pE_div_ZZ_p "div"( ZZ_pE_c x, ZZ_pE_c a, ZZ_p_c b) # raises an error if b not invertible
    void ZZ_pE_ZZ_p_div "div"( ZZ_pE_c x, ZZ_p_c a, ZZ_pE_c b) # raises an error if b not invertible
    void ZZ_pE_random "random"( ZZ_pE_c x)
    void ZZ_pE_trace "trace"( ZZ_p_c x, ZZ_pE_c a)
    void ZZ_pE_norm "norm"( ZZ_p_c x, ZZ_pE_c a)

    ZZ_pE_c long_to_ZZ_pE "to_ZZ_pE"(long i)
    ZZ_pE_c ZZ_to_ZZ_pE "to_ZZ_pE"(ZZ_c i)
    ZZ_pE_c ZZ_p_to_ZZ_pE "to_ZZ_pE"(ZZ_p_c i)
    ZZ_pE_c ZZ_pX_to_ZZ_pE "to_ZZ_pE"(ZZ_pX_c i)


cdef extern from "ntlwrap_impl.h":
    ZZ_pX_c ZZ_pE_to_ZZ_pX(ZZ_pE_c x)



# sage_setup: distribution = sagemath-ntl
r"""
Linkage for arithmetic with NTL's GF2X elements.

This file provides the backend for \class{Polynomial_GF2X} via
templating.

AUTHOR:
    -- Martin Albrecht (2008-10): initial version
"""

#*****************************************************************************
#       Copyright (C) 2008 Martin Albrecht <M.R.Albrecht@rhul.ac.uk>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  http://www.gnu.org/licenses/
#*****************************************************************************

from cysignals.signals cimport sig_on, sig_off

from sage.libs.ntl.GF2 cimport *
from sage.libs.ntl.GF2X cimport *


cdef GF2X_c *celement_new(long parent) noexcept:
    """
    EXAMPLES::

        sage: P.<x> = GF(2)[]
    """
    return new GF2X_c()

cdef int celement_delete(GF2X_c *e, long parent) noexcept:
    """
    EXAMPLES::

        sage: P.<x> = GF(2)[]
        sage: del x
    """
    del e

cdef int celement_construct(GF2X_c *e, long parent) noexcept:
    """
    EXAMPLES::

        sage: P.<x> = GF(2)[]
    """
    pass

cdef int celement_destruct(GF2X_c *e, long parent) noexcept:
    """
    EXAMPLES::

        sage: P.<x> = GF(2)[]
        sage: del x
    """
    pass

cdef int celement_gen(GF2X_c *e, long i, long parent) except -2:
    """
    EXAMPLES::

        sage: P.<x> = GF(2)[]
    """
    cdef unsigned char g = 2
    GF2XFromBytes(e[0], <unsigned char *>(&g), 1)

cdef object celement_repr(GF2X_c *e, long parent):
    """
    We ignore NTL's printing.

    EXAMPLES::

        sage: P.<x> = GF(2)[]
        sage: x
        x
    """
    raise NotImplementedError

cdef inline int celement_set(GF2X_c* res, GF2X_c* a, long parent) except -2:
    """
    EXAMPLES::

        sage: P.<x> = GF(2)[]
        sage: y = x; y
        x
    """
    res[0] = a[0]

cdef inline int celement_set_si(GF2X_c* res, long i, long parent) except -2:
    """
    EXAMPLES::

        sage: P.<x> = GF(2)[]
        sage: P(0)
        0
        sage: P(2)
        0
        sage: P(1)
        1
    """
    GF2X_conv_long(res[0], i)

cdef inline long celement_get_si(GF2X_c* res, long parent) except -2:
    raise NotImplementedError

cdef inline bint celement_is_zero(GF2X_c* a, long parent) except -2:
    """
    EXAMPLES::

        sage: P.<x> = GF(2)[]
        sage: bool(x), x.is_zero()
        (True, False)
        sage: bool(P(0)), P(0).is_zero()
        (False, True)
    """
    return GF2X_IsZero(a[0])

cdef inline bint celement_is_one(GF2X_c *a, long parent) except -2:
    """
    EXAMPLES::

        sage: P.<x> = GF(2)[]
        sage: x.is_one()
        False
        sage: P(1).is_one()
        True
    """
    return GF2X_IsOne(a[0])

cdef inline bint celement_equal(GF2X_c *a, GF2X_c *b, long parent) except -2:
    """
    EXAMPLES::

        sage: P.<x> = GF(2)[]
        sage: x == x
        True
        sage: y = x; x == y
        True
        sage: x^2 + 1 == x^2 + x
        False
    """
    return a[0] == b[0]

cdef inline int celement_cmp(GF2X_c *a, GF2X_c *b, long parent) except -2:
    """
    EXAMPLES::

        sage: P.<x> = GF(2)[]
        sage: x != 1
        True
        sage: x < 1
        False
        sage: x > 1
        True

        sage: f = x^64 + x^20 + 1
        sage: g = x^63 + x^20 + 1
        sage: f > g
        True

        sage: f = x^64 + x^10 + 1
        sage: g = x^64 + x^20 + 1
        sage: f < g
        True

        sage: f = x^64 + x^20
        sage: g = x^64 + x^20 + 1
        sage: f < g
        True
    """
    cdef bint t
    cdef long diff
    cdef long ca, cb
    diff = GF2X_NumBits(a[0]) - GF2X_NumBits(b[0])
    if diff > 0:
        return 1
    elif diff < 0:
        return -1
    else:
        for i in range(GF2X_NumBits(a[0])-1, -1, -1):
            ca = GF2_conv_to_long(GF2X_coeff(a[0], i))
            cb = GF2_conv_to_long(GF2X_coeff(b[0], i))
            if ca < cb:
                return -1
            elif ca > cb:
                return 1
    return 0

cdef long celement_len(GF2X_c *a, long parent) except -2:
    """
    EXAMPLES::

        sage: P.<x> = GF(2)[]
        sage: x.degree()
        1
        sage: (x+1).degree()
        1
    """
    return int(GF2X_NumBits(a[0]))

cdef inline int celement_add(GF2X_c *res, GF2X_c *a, GF2X_c *b, long parent) except -2:
    """
    EXAMPLES::

        sage: P.<x> = GF(2)[]
        sage: x + 1
        x + 1
    """
    GF2X_add(res[0], a[0], b[0])

cdef inline int celement_sub(GF2X_c* res, GF2X_c* a, GF2X_c* b, long parent) except -2:
    """
    EXAMPLES::

        sage: P.<x> = GF(2)[]
        sage: x - 1
        x + 1
    """
    GF2X_sub(res[0], a[0], b[0])

cdef inline int celement_neg(GF2X_c* res, GF2X_c* a, long parent) except -2:
    """
    EXAMPLES::

        sage: P.<x> = GF(2)[]
        sage: -x
        x
    """
    res[0] = a[0]

cdef inline int celement_mul_scalar(GF2X_c* res, GF2X_c* p, object c,
        long parent) except -1:
    """
    TESTS::

        sage: P.<x> = GF(2)[]
        sage: p = P.random_element()
        sage: 0*p
        0
        sage: 1*p == p
        True
        sage: (3^97)*p == p
        True
    """
    if int(c) == 0:
        GF2X_conv_long(res[0], 0)
    else:
        res[0] = p[0]

cdef inline int celement_mul(GF2X_c* res, GF2X_c* a, GF2X_c* b, long parent) except -2:
    """
    EXAMPLES::

        sage: P.<x> = GF(2)[]
        sage: x*(x+1)
        x^2 + x
    """
    GF2X_mul(res[0], a[0], b[0])

cdef inline int celement_truncate(GF2X_c* res, GF2X_c* a, long len, long parent) except -2:
    """
    EXAMPLES::

        sage: P.<x> = GF(2)[]
        sage: p = x^5 + x^3 + x^2 + x + 1
        sage: p.truncate(2)   # indirect doctest
        x + 1
    """
    GF2X_trunc(res[0], a[0], len)

cdef inline int celement_div(GF2X_c* res, GF2X_c* a, GF2X_c* b, long parent) except -2:
    """
    EXAMPLES::

        sage: P.<x> = GF(2)[]
    """
    return GF2X_divide(res[0], a[0], b[0])

cdef inline int celement_floordiv(GF2X_c* res, GF2X_c* a, GF2X_c* b, long parent) except -2:
    """
    EXAMPLES::

        sage: P.<x> = GF(2)[]
        sage: x//(x + 1)
        1
        sage: (x + 1)//x
        1
    """
    GF2X_div(res[0], a[0], b[0])

cdef inline int celement_mod(GF2X_c* res, GF2X_c* a, GF2X_c* b, long parent) except -2:
    """
    EXAMPLES::

        sage: P.<x> = GF(2)[]
        sage: (x^2 + 1) % x^2
        1
    """
    GF2X_rem(res[0], a[0], b[0])

cdef inline int celement_quorem(GF2X_c* q, GF2X_c* r, GF2X_c* a, GF2X_c* b, long parent) except -2:
    """
    EXAMPLES::

        sage: P.<x> = GF(2)[]
        sage: f = x^2 + x + 1
        sage: f.quo_rem(x + 1)
        (x, 1)
    """
    GF2X_DivRem(q[0], r[0], a[0], b[0])

cdef inline int celement_inv(GF2X_c* res, GF2X_c* a, long parent) except -2:
    """
    We ignore NTL here and use the fraction field constructor.

    EXAMPLES::

        sage: P.<x> = GF(2)[]
    """
    raise NotImplementedError

cdef inline int celement_pow(GF2X_c* res, GF2X_c* x, long e, GF2X_c *modulus, long parent) except -2:
    """
    EXAMPLES::

        sage: P.<x> = GF(2)[]
        sage: x^1000
        x^1000
        sage: (x+1)^2
        x^2 + 1
        sage: (x+1)^(-2)
        1/(x^2 + 1)
        sage: f = x^9 + x^7 + x^6 + x^5 + x^4 + x^2 + x
        sage: h = x^10 + x^7 + x^6 + x^5 + x^4 + x^3 + x^2 + 1
        sage: (f^2) % h
        x^9 + x^8 + x^7 + x^5 + x^3
        sage: pow(f, 2, h)
        x^9 + x^8 + x^7 + x^5 + x^3
        sage: pow(x, 1000, h)
        x^8 + x^7 + x^4

    Check that deg x >= deg modulus works (:issue:`35324`)::

        sage: pow(x+1, 2, x^2+x+1)
        x
        sage: pow(x^2+1, 2, x^2+x+1)
        x + 1
    """
    cdef GF2XModulus_c mod
    cdef GF2X_c xmod

    if modulus == NULL:
        if GF2X_IsX(x[0]):
            GF2X_LeftShift(res[0], x[0], e - 1)
        else:
            do_sig = GF2X_deg(x[0]) > 1e5
            if do_sig:
                sig_on()
            GF2X_power(res[0], x[0], e)
            if do_sig:
                sig_off()
    else:
        GF2X_rem(xmod, x[0], modulus[0])
        GF2XModulus_build(mod, modulus[0])
        if GF2X_IsX(xmod):
            GF2X_PowerXMod_long_pre(res[0], e, mod)
        else:
            do_sig = GF2X_deg(x[0]) > 1e5
            if do_sig:
                sig_on()
            GF2X_PowerMod_long_pre(res[0], xmod, e, mod)
            if do_sig:
                sig_off()


cdef inline int celement_gcd(GF2X_c* res, GF2X_c* a, GF2X_c *b, long parent) except -2:
    """
    EXAMPLES::

        sage: P.<x> = GF(2)[]
        sage: f = x*(x+1)
        sage: f.gcd(x+1)
        x + 1
        sage: f.gcd(x^2)
        x
    """
    GF2X_GCD(res[0], a[0], b[0])

cdef inline int celement_xgcd(GF2X_c* res, GF2X_c* s, GF2X_c *t, GF2X_c* a, GF2X_c *b, long parent) except -2:
    """
    EXAMPLES::

        sage: P.<x> = GF(2)[]
        sage: f = x*(x+1)
        sage: f.xgcd(x+1)
        (x + 1, 0, 1)
        sage: f.xgcd(x^2)
        (x, 1, 1)
    """
    GF2X_XGCD(res[0], s[0], t[0], a[0], b[0])



# sage_setup: distribution = sagemath-ntl
ctypedef double celement

include "matrix_modn_dense_template_header.pxi"

from sage.rings.finite_rings.integer_mod cimport IntegerMod_abstract

cdef class Matrix_modn_dense_double(Matrix_modn_dense_template):
    cdef IntegerMod_abstract _get_template
    cdef bint _fits_int32



# sage_setup: distribution = sagemath-linbox
# distutils: language = c++
# distutils: libraries = CBLAS_LIBRARIES
# distutils: library_dirs = CBLAS_LIBDIR
# distutils: include_dirs = CBLAS_INCDIR
# distutils: extra_compile_args = -D_XPG6
# sage.doctest: needs sage.rings.finite_rings
r"""
Dense matrices over `\ZZ/n\ZZ` for `n < 94906266` using LinBox's ``Modular<double>``

AUTHORS:

- Burcin Erocal
- Martin Albrecht
"""
# #############################################################################
#       Copyright (C) 2011 Burcin Erocal <burcin@erocal.org>
#       Copyright (C) 2011 Martin Albrecht <martinralbrecht@googlemail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************
from sage.rings.finite_rings.stdint cimport *

from sage.libs.linbox.givaro cimport \
    Modular_double as ModField, \
    Poly1Dom, Dense

from sage.libs.linbox.linbox cimport \
    reducedRowEchelonize, \
    DenseMatrix_Modular_double as DenseMatrix

from sage.libs.linbox.fflas cimport \
    fgemm, pfgemm, fgemv, Det, pDet, Rank, pRank, ReducedRowEchelonForm, pReducedRowEchelonForm, applyP, \
    MinPoly, CharPoly, MinPoly, \
    ModDoubleDensePolynomial as ModDensePoly

ctypedef Poly1Dom[ModField, Dense] ModDensePolyRing

# Limit for LinBox Modular<double>
MAX_MODULUS = 94906266

from sage.rings.finite_rings.integer_mod cimport IntegerMod_int64

include "matrix_modn_dense_template.pxi"


cdef class Matrix_modn_dense_double(Matrix_modn_dense_template):
    r"""
    Dense matrices over `\ZZ/n\ZZ` for `n < 94906266` using LinBox's ``Modular<double>``.

    These are matrices with integer entries mod ``n`` represented as
    floating-point numbers in a 64-bit word for use with LinBox routines.
    This allows for ``n`` up to `94906266`. By default, the analogous
    ``Matrix_modn_dense_float`` class is used for smaller moduli, specifically
    for ``n`` up to `2^{8}`.

    Routines here are for the most basic access, see the
    ``matrix_modn_dense_template.pxi`` file for higher-level routines.
    """

    def __cinit__(self):
        """
        The Cython constructor.

        TESTS::

            sage: A = random_matrix(IntegerModRing(2^16), 4, 4)
            sage: type(A[0,0])
            <class 'sage.rings.finite_rings.integer_mod.IntegerMod_int64'>
        """
        self._get_template = self._base_ring.zero()
        # note that INTEGER_MOD_INT32_LIMIT is ceil(sqrt(2^31-1)) < 94906266
        self._fits_int32 = ((<Matrix_modn_dense_template>self).p <= INTEGER_MOD_INT32_LIMIT)

    cdef void set_unsafe_int(self, Py_ssize_t i, Py_ssize_t j, int value) noexcept:
        r"""
        Set the (i,j) entry of ``self`` to the int value.

        EXAMPLES::

            sage: A = random_matrix(GF(3016963), 4, 4)
            sage: l = A.list()
            sage: A[0,0] = 220082r
            sage: A.list()[1:] == l[1:]
            True
            sage: a = A[0,0]; a
            220082
            sage: ~a
            2859358

            sage: A = random_matrix(Integers(5099106), 4, 4)
            sage: l = A.list()
            sage: A[0,0] = 220082r
            sage: A.list()[1:] == l[1:]
            True
            sage: a = A[0,0]; a
            220082
            sage: a*a
            4777936
        """
        self._matrix[i][j] = <double>value

    cdef set_unsafe(self, Py_ssize_t i, Py_ssize_t j, x):
        r"""
        Set the (i,j) entry with no bounds-checking, or any other checks.

        Assumes that `x` is in the base ring.

        EXAMPLES::

            sage: A = random_matrix(GF(3016963), 4, 4)
            sage: K = A.base_ring()
            sage: l = A.list()
            sage: A[0,0] = K(220082)
            sage: A.list()[1:] == l[1:]
            True
            sage: a = A[0,0]; a
            220082
            sage: ~a
            2859358

            sage: A = random_matrix(Integers(5099106), 4, 4)
            sage: K = A.base_ring()
            sage: l = A.list()
            sage: A[0,0] = K(220081)
            sage: A.list()[1:] == l[1:]
            True
            sage: a = A[0,0]; a
            220081
            sage: a*a
            4337773
        """
        if (<Matrix_modn_dense_double>self)._fits_int32:
            self._matrix[i][j] = <double>(<IntegerMod_int>x).ivalue
        else:
            self._matrix[i][j] = <double>(<IntegerMod_int64>x).ivalue

    cdef IntegerMod_abstract get_unsafe(self, Py_ssize_t i, Py_ssize_t j):
        r"""
        Return the (i,j) entry with no bounds-checking.

        OUTPUT:

        A :class:`sage.rings.finite_rings.integer_mod.IntegerMod_int`
        or
        :class:`sage.rings.finite_rings.integer_mod.IntegerMod_int64`
        object, depending on the modulus.

        EXAMPLES::

            sage: K = GF(3016963)
            sage: l = [K.random_element() for _ in range(4*4)]
            sage: A = matrix(K, 4, 4, l)
            sage: a = A[0,0]
            sage: a == l[0]
            True
            sage: a == 0 or ~a*a == 1
            True
            sage: a.parent() is K
            True

            sage: K = Integers(5099106)
            sage: l = [K.random_element() for _ in range(4*4)]
            sage: A = matrix(Integers(5099106), 4, 4, l)
            sage: a = A[0,1]
            sage: a == l[1]
            True
            sage: a*a == K(Integer(l[1]))^2
            True
        """
        cdef Matrix_modn_dense_double _self = <Matrix_modn_dense_double>self
        cdef double result = (<Matrix_modn_dense_template>self)._matrix[i][j]
        if _self._fits_int32:
            return (<IntegerMod_int>_self._get_template)._new_c(<int_fast32_t>result)
        else:
            return (<IntegerMod_int64>_self._get_template)._new_c(<int_fast64_t>result)

    cdef copy_from_unsafe(self, Py_ssize_t iDst, Py_ssize_t jDst, src, Py_ssize_t iSrc, Py_ssize_t jSrc):
        r"""
        Copy the ``(iSrc, jSrc)`` entry of ``src`` into the ``(iDst, jDst)``
        entry of ``self``.

        INPUT:

        - ``iDst`` - the row to be copied to in ``self``.
        - ``jDst`` - the column to be copied to in ``self``.
        - ``src`` - the matrix to copy from. Should be a
                    Matrix_modn_dense_double with the same base ring as
                    ``self``.
        - ``iSrc``  - the row to be copied from in ``src``.
        - ``jSrc`` - the column to be copied from in ``src``.

        TESTS::

            sage: m = matrix(GF(257),3,4,range(12))
            sage: m
            [ 0  1  2  3]
            [ 4  5  6  7]
            [ 8  9 10 11]
            sage: m.transpose()
            [ 0  4  8]
            [ 1  5  9]
            [ 2  6 10]
            [ 3  7 11]
            sage: m.matrix_from_rows([0,2])
            [ 0  1  2  3]
            [ 8  9 10 11]
            sage: m.matrix_from_columns([1,3])
            [ 1  3]
            [ 5  7]
            [ 9 11]
            sage: m.matrix_from_rows_and_columns([1,2],[0,3])
            [ 4  7]
            [ 8 11]
        """
        cdef Matrix_modn_dense_double _src = <Matrix_modn_dense_double>src
        self._matrix[iDst][jDst] = _src._matrix[iSrc][jSrc]
        


# sage_setup: distribution = sagemath-ntl
from sage.libs.ntl.types cimport GF2E_c
from sage.libs.ntl.ntl_GF2EContext cimport ntl_GF2EContext_class

cdef class ntl_GF2E():
    cdef GF2E_c x
    cdef ntl_GF2EContext_class c
    cdef ntl_GF2E _new(self)



# sage_setup: distribution = sagemath-ntl
from sage.libs.ntl.types cimport mat_GF2_c, vec_GF2_c, GF2_c


cdef extern from "ntlwrap.h":
    void mat_GF2_add "add"( mat_GF2_c x, mat_GF2_c a, mat_GF2_c b)
    void mat_GF2_sub "sub"( mat_GF2_c x, mat_GF2_c a, mat_GF2_c b)
    void mat_GF2_mul "mul"( mat_GF2_c x, mat_GF2_c a, mat_GF2_c b)
    void mat_GF2_negate "NTL::negate"(mat_GF2_c x, mat_GF2_c a)
    void mat_GF2_power "NTL::power"(mat_GF2_c t, mat_GF2_c x, long e)
    GF2_c mat_GF2_determinant "determinant"(mat_GF2_c m)
    void mat_GF2_transpose "transpose"(mat_GF2_c r, mat_GF2_c m)
    long mat_GF2_IsZero "IsZero"(mat_GF2_c x)

    long mat_GF2_gauss "gauss"(mat_GF2_c A, long w)
    void mat_GF2_solve "solve"(GF2_c d, vec_GF2_c X, mat_GF2_c A, vec_GF2_c b)
    void mat_GF2_inv "inv" (mat_GF2_c X, mat_GF2_c A)

    long mat_GF2_IsIdent "IsIdent"(mat_GF2_c A, long n)
    long mat_GF2_IsDiag "IsDiag"(mat_GF2_c A, long n, GF2_c d)

    void mat_GF2_image "image"(mat_GF2_c X, mat_GF2_c A)
    void mat_GF2_kernel "kernel" (mat_GF2_c X, mat_GF2_c A)

    void vec_GF2_conv_mat_GF2 "conv" (vec_GF2_c out, mat_GF2_c inp)
    void mat_GF2_conv_vec_GF2(mat_GF2_c out, vec_GF2_c inp)


cdef extern from "ntlwrap_impl.h":
    void mat_GF2_setitem(mat_GF2_c* x, int i, int j, GF2_c* z)



# sage_setup: distribution = sagemath-ntl
from sage.libs.ntl.types cimport *
from sage.libs.ntl.ZZ cimport *
from sage.libs.ntl.ZZ_pX cimport *
from sage.libs.ntl.ZZ_p cimport *
from sage.libs.ntl.ZZX cimport *
from sage.libs.ntl.ZZ_pE cimport *
from sage.libs.ntl.ZZ_pEX cimport *
from sage.libs.ntl.GF2 cimport *
from sage.libs.ntl.GF2X cimport *
from sage.libs.ntl.GF2E cimport *
from sage.libs.ntl.GF2EX cimport *
from sage.libs.ntl.vec_GF2 cimport *
from sage.libs.ntl.vec_GF2E cimport *
from sage.libs.ntl.mat_ZZ cimport *
from sage.libs.ntl.mat_GF2 cimport *
from sage.libs.ntl.mat_GF2E cimport *
from sage.libs.ntl.convert cimport mpz_to_ZZ, ZZ_to_mpz



# sage_setup: distribution = sagemath-linbox
# distutils: extra_compile_args = GIVARO_CFLAGS
# distutils: include_dirs = GIVARO_INCDIR
# distutils: libraries = GIVARO_LIBRARIES FFLASFFPACK_LIBRARIES
# distutils: library_dirs = GIVARO_LIBDIR
# distutils: language = c++

from libc.stdint cimport uint32_t, uint64_t

from sage.libs.gmp.types cimport (mpz_t, mpz_srcptr, mpz_ptr,
                                  mpq_t, mpq_srcptr, mpq_ptr)

cdef extern from "<iostream>" namespace "std":
    cdef cppclass ostream:
        ostream& write(const char*, int) except +

cdef extern from "gmp++/gmp++.h" namespace "Givaro":
    cdef cppclass Integer:
        Integer()
        Integer(int32_t)
        Integer(int64_t)
        Integer(uint32_t)
        Integer(uint64_t)
        Integer(Integer&)

        mpz_ptr get_mpz()
        mpz_srcptr get_mpz_const()

cdef extern from "givaro/givrational.h" namespace "Givaro":
    cdef cppclass Rational:
        Rational()
        Rational(int32_t)
        Rational(int64_t)
        Rational(uint32_t)
        Rational(uint64_t)
        Rational(Integer&)
        Rational(int32_t, int32_t)
        Rational(int64_t, int64_t)
        Rational(uint32_t, uint32_t)
        Rational(uint64_t, uint64_t)
        Rational(Integer&, Integer&, int)

        Integer nume()
        Integer deno()

cdef extern from "givaro/givcategory.h" namespace "Givaro":
    cdef cppclass Sporadic:
        pass
    cdef cppclass Dense:
        pass
    cdef cppclass Sparse:
        pass

cdef extern from "givaro/zring.h":
    ## template<class _Element> class ZRing
    cdef cppclass ZRing "Givaro::ZRing<Givaro::Integer>":
        ctypedef Integer Element
        Element zero
        Element one
        Element mone

cdef extern from "givaro/qfield.h":
    ## template<class RatElement> class QField
    cdef cppclass QField "Givaro::QField<Givaro::Rational>":
        ctypedef Rational Element
        Element one
        Element mOne
        Element zero

cdef extern from "givaro/modular-integral.h":
    cdef cppclass Modular_uint64 "Givaro::Modular<uint64_t>":
        ctypedef uint64_t Element
        Modular_uint64(int modulus)

        Element init(Element res, int v)
        Element inv(Element x, Element y)
        Element neg(Element x, Element y)
        Element mul(Element r, Element x, Element y)
        Element mulin(Element x, Element y)
        Element addin(Element x, Element y)
        Element invin(Element y)
        Element negin(Element y)
        int characteristic(int c)
        bint isZero(Element x)

        ostream& write(ostream&)

cdef extern from "givaro/modular-floating.h":
    cdef cppclass Modular_double "Givaro::Modular<double>":
        ctypedef double Element
        Modular_double(int modulus)

        Element init(Element res, int v)
        Element init(Element res, double v)
        Element inv(Element x, Element y)
        Element neg(Element x, Element y)
        Element mul(Element r, Element x, Element y)
        Element mulin(Element x, Element y)
        Element addin(Element x, Element y)
        Element invin(Element y)
        Element negin(Element y)
        int characteristic(int c)
        bint isZero(Element x)

        ostream& write(ostream&)

    cdef cppclass Modular_float "Givaro::Modular<float>":
        ctypedef float Element
        Modular_float(int modulus)

        Element init(Element res, int v)
        Element init(Element res, float v)
        Element inv(Element x, Element y)
        Element neg(Element x, Element y)
        Element mul(Element r, Element x, Element y)
        Element mulin(Element x, Element y)
        Element addin(Element x, Element y)
        Element invin(Element y)
        Element negin(Element y)
        int characteristic(int c)
        bint isZero(Element x)

        ostream& write(ostream&)

cdef extern from "givaro/givpoly1.h" namespace "Givaro":
    ## template < typename T, typename A=std::allocator<T> >
    ## class givvector : public __GIV_STANDARD_VECTOR<T,A>
    cdef cppclass givvector [T,ALLOCATOR=*]:
        T& operator[](size_t i)
        size_t size()

    ## template<class Domain, class StorageTag=Dense> class Poly1Dom
    cdef cppclass Poly1Dom[Domain,StorageClass=*]:
        Poly1Dom(Domain&)



# sage_setup: distribution = sagemath-ntl
from sage.libs.ntl.types cimport GF2_c


cdef extern from "ntlwrap.h":
    int GF2_IsOne "IsOne"(GF2_c x)
    int GF2_IsZero "IsZero"(GF2_c x)

    void GF2_add "add"( GF2_c x, GF2_c a, GF2_c b)
    void GF2_sub "sub"( GF2_c x, GF2_c a, GF2_c b)
    void GF2_mul "mul"( GF2_c x, GF2_c a, GF2_c b)
    void GF2_div "div"( GF2_c x, GF2_c a, GF2_c b)
    void GF2_negate "NTL::negate"(GF2_c x, GF2_c a)
    void GF2_power "NTL::power"(GF2_c t, GF2_c x, long e)
    long GF2_deg "deg"(GF2_c x)

    void GF2_conv_long "conv" (GF2_c x, long i)
    long GF2_conv_to_long "rep" (GF2_c x)



# sage_setup: distribution = sagemath-ntl
from sage.libs.ntl.types cimport ZZ_p_c
from sage.libs.ntl.ntl_ZZ_pContext cimport ntl_ZZ_pContext_class

cdef class ntl_ZZ_p():
    cdef ZZ_p_c x
    cdef ntl_ZZ_pContext_class c
    cdef int get_as_int(ntl_ZZ_p self) noexcept
    cdef void set_from_int(ntl_ZZ_p self, int value) noexcept
    cdef ntl_ZZ_p _new(self)



# sage_setup: distribution = sagemath-ntl
from sage.libs.ntl.types cimport mat_GF2_c
from sage.libs.ntl.ntl_GF2 cimport ntl_GF2

cdef class ntl_mat_GF2():
    cdef mat_GF2_c x
    cdef ntl_GF2 _new_element(self)
    cdef ntl_mat_GF2 _new(self)



# sage_setup: distribution = sagemath-ntl
from sage.libs.ntl.types cimport zz_pContext_c

cdef class ntl_zz_pContext_class():
    cdef zz_pContext_c x
    cdef void restore_c(self) noexcept
    cdef long p



# sage_setup: distribution = sagemath-ntl
from sage.libs.ntl.types cimport mat_GF2E_c
from sage.libs.ntl.ntl_GF2EContext cimport ntl_GF2EContext_class
from sage.libs.ntl.ntl_GF2E cimport ntl_GF2E

cdef class ntl_mat_GF2E():
    cdef mat_GF2E_c x
    cdef ntl_GF2EContext_class c
    cdef ntl_GF2E _new_element(self)
    cdef ntl_mat_GF2E _new(self)



# sage_setup: distribution = sagemath-ntl
from sage.libs.ntl.lzz_p cimport *
from sage.libs.ntl.lzz_pX cimport *

from sage.libs.ntl.ntl_lzz_pContext cimport ntl_zz_pContext_class

cdef class ntl_zz_pX():
    cdef zz_pX_c x
    cdef ntl_zz_pContext_class c
    cdef ntl_zz_pX _new(self)



# sage_setup: distribution = sagemath-ntl
# distutils: depends = NTL/ZZ.h

from sage.libs.ntl.types cimport zz_p_c

cdef extern from "ntlwrap.h":
    long zz_p_rep "rep"(zz_p_c x)
    long zz_p_isZero "IsZero"(zz_p_c x)
    void zz_p_add "add"(zz_p_c x, zz_p_c a, zz_p_c b)
    void zz_p_sub "sub"(zz_p_c x, zz_p_c a, zz_p_c b)
    void zz_p_negate "NTL::negate"(zz_p_c x, zz_p_c a)
    void zz_p_mul "mul"(zz_p_c x, zz_p_c a, zz_p_c b)
    void zz_p_div "div"(zz_p_c x, zz_p_c a, zz_p_c b)
    void zz_p_inv "inv"(zz_p_c x, zz_p_c a)
    void zz_p_power "NTL::power"(zz_p_c x, zz_p_c a, long e)
    void zz_p_sqr "sqr"(zz_p_c x, zz_p_c a)
    void zz_p_clear "clear"(zz_p_c x)
    void zz_p_set_zero "clear"(zz_p_c x)
    void zz_p_set_one "set"(zz_p_c x)
    void zz_p_swap "swap"(zz_p_c x, zz_p_c y)



# sage_setup: distribution = sagemath-linbox
# sage.doctest: needs sage.libs.linbox

from libc.stdint cimport uint64_t
from libc.limits cimport UINT_MAX

cimport sage.libs.linbox.givaro as givaro
cimport sage.libs.linbox.linbox as linbox

from sage.arith.misc import is_prime
from sage.libs.linbox.conversion cimport (METHOD_DEFAULT,
                                          METHOD_DENSE_ELIMINATION,
                                          METHOD_SPARSE_ELIMINATION,
                                          METHOD_BLACKBOX,
                                          METHOD_WIEDEMANN,
                                          new_linbox_matrix_modn_sparse)
from sage.matrix.matrix_modn_sparse cimport Matrix_modn_sparse


def _rank_det_linbox(Matrix_modn_sparse self):
    """
    Return the rank and determinant using linbox.

    .. NOTE::

        This method does not perform any caching contrarily to
        :meth:`determinant` and :meth:`rank`.

    EXAMPLES::

        sage: from sage.matrix.matrix_modn_sparse_linbox import _rank_det_linbox
        sage: m = matrix(Zmod(13), 1, sparse=True)
        sage: m[0,0] = 0
        sage: m._rank_det_linbox()
        (0, 0)
        sage: for i in range(1, 13):
        ....:     m[0,0] = i
        ....:     assert m._rank_det_linbox() == (1, i)

        sage: m = matrix(GF(5), 2, sparse=True)
        sage: m[0,0] = 1
        sage: m[0,1] = 2
        sage: m[1,0] = 1
        sage: m[1,1] = 3
        sage: m._rank_det_linbox()
        (2, 1)
        sage: m
        [1 2]
        [1 3]

    TESTS::

        sage: matrix(Zmod(3), 0, sparse=True)._rank_det_linbox()
        (0, 1)
    """
    if self._nrows == 0 or self._ncols == 0:
        # TODO: bug in linbox (gives segfault)
        return 0, self.base_ring().one()

    cdef size_t A_rank = 0
    cdef uint64_t A_det = 0

    if not is_prime(self.p):
        raise TypeError("only GF(p) supported via LinBox")

    cdef givaro.Modular_uint64 * F = new givaro.Modular_uint64(<uint64_t> self.p)
    cdef linbox.SparseMatrix_Modular_uint64 * A
    A = new_linbox_matrix_modn_sparse(F[0], self)

    cdef linbox.GaussDomain_Modular_uint64 * dom = new linbox.GaussDomain_Modular_uint64(F[0])

    # NOTE: see above for the reason of casting...
    if A.rowdim() >= <size_t> UINT_MAX or A.coldim() >= <size_t> UINT_MAX:
        raise ValueError("row/column size unsupported in LinBox")

    dom.InPlaceLinearPivoting(A_rank, A_det, A[0], A.rowdim(), A.coldim())

    del A
    del F
    del dom

    return <long> A_rank, self.base_ring()(A_det)



# sage_setup: distribution = sagemath-ntl
# distutils: depends = NTL/ZZ.h

from sage.libs.ntl.types cimport ZZ_c


cdef extern from "ntlwrap.h":
    void ZZ_conv_from_int "conv"(ZZ_c x, int i)
    void ZZ_conv_to_int "conv"(int i, ZZ_c x)
    void ZZ_conv_from_long "conv"(ZZ_c x, long l)
    void ZZ_conv_to_long "conv"(long l, ZZ_c x)

    long ZZ_sign "sign"(ZZ_c a)
    long ZZ_IsZero "IsZero"(ZZ_c a)
    long ZZ_IsOne "IsOne"(ZZ_c a)
    long ZZ_compare "compare"(ZZ_c a, ZZ_c b)
    void ZZ_add "add"( ZZ_c x, ZZ_c a, ZZ_c b)
    void ZZ_add_long "add"(ZZ_c x, ZZ_c a, long b)
    void ZZ_sub "sub"( ZZ_c x, ZZ_c a, ZZ_c b)
    void ZZ_sub_long "sub"(ZZ_c x, long a, ZZ_c b)
    void ZZ_negate "NTL::negate"(ZZ_c x, ZZ_c a)
    void ZZ_abs "abs"(ZZ_c x, ZZ_c a)

    void ZZ_mul "mul"( ZZ_c x, ZZ_c a, ZZ_c b)
    void ZZ_mul_long "mul"( ZZ_c x, ZZ_c a, long b)
    void ZZ_sqr "sqr"(ZZ_c x, ZZ_c a)

    void ZZ_DivRem "DivRem"(ZZ_c q, ZZ_c r, ZZ_c a, ZZ_c b)
    void ZZ_div "div"( ZZ_c x, ZZ_c a, ZZ_c b)
    void ZZ_rem "rem"( ZZ_c r, ZZ_c a, ZZ_c b)
    long ZZ_DivRem_long "DivRem"(ZZ_c q, ZZ_c a, long b)
    long ZZ_rem_long "rem"(ZZ_c a, long b)
    long ZZ_divide "divide"(ZZ_c q, ZZ_c a, ZZ_c b)
    long ZZ_divide_long "divide"(ZZ_c q, ZZ_c a, long b)
    long ZZ_divide_test "divide"(ZZ_c a, ZZ_c b)
    long ZZ_divide_test_long "divide"(ZZ_c a, long b)

    void ZZ_GCD "GCD"(ZZ_c d, ZZ_c a, ZZ_c b)
    void ZZ_XGCD "XGCD"(ZZ_c d, ZZ_c s, ZZ_c t, ZZ_c a, ZZ_c b)

    void ZZ_AddMod "AddMod"(ZZ_c x, ZZ_c a, ZZ_c b, ZZ_c n)
    void ZZ_SubMod "SubMod"(ZZ_c x, ZZ_c a, ZZ_c b, ZZ_c n)
    void ZZ_NegateMod "NegateMod"(ZZ_c x, ZZ_c a, ZZ_c n)
    void ZZ_MulMod "MulMod"(ZZ_c x, ZZ_c a, ZZ_c b, ZZ_c n)
    void ZZ_SqrMod "SqrMod"(ZZ_c x, ZZ_c a, ZZ_c n)
    void ZZ_InvMod "InvMod"(ZZ_c x, ZZ_c a, ZZ_c n)
    long ZZ_InvModStatus "InvModStatus"(ZZ_c x, ZZ_c a, ZZ_c n)
    void ZZ_PowerMod "PowerMod"(ZZ_c x, ZZ_c a, ZZ_c e, ZZ_c n)
    void ZZ_PowerMod_long "PowerMod"(ZZ_c x, ZZ_c a, long e, ZZ_c n)

    void ZZ_LeftShift "LeftShift"(ZZ_c x, ZZ_c a, long n)
    void ZZ_RightShift "RightShift"(ZZ_c x, ZZ_c a, long n)

    long ZZ_MakeOdd "MakeOdd"(ZZ_c x)
    long ZZ_NumTwos "NumTwos"(ZZ_c x)
    long ZZ_IsOdd "IsOdd"(ZZ_c a)
    long ZZ_NumBits "NumBits"(ZZ_c a)
    long ZZ_bit "bit"(ZZ_c a, long k)
    void ZZ_trunc "trunc"(ZZ_c x, ZZ_c a, long k)
    void ZZ_SetBit "SetBit"(ZZ_c x, long p)
    void ZZ_SwitchBit "SwitchBit"(ZZ_c x, long p)
    void ZZ_weight "weight"(ZZ_c a)
    void ZZ_bit_and "bit_and"(ZZ_c x, ZZ_c a, ZZ_c b)
    void ZZ_bit_or "bit_or"(ZZ_c x, ZZ_c a, ZZ_c b)
    void ZZ_bit_xor "bit_xor"(ZZ_c x, ZZ_c a, ZZ_c b)

    void ZZ_SetSeed "SetSeed"(ZZ_c s)
    void ZZ_RandomBnd "RandomBnd"(ZZ_c x, ZZ_c n)
    void ZZ_RandomBits "RandomBits"(ZZ_c x, long l)

    long ZZ_CRT "CRT"(ZZ_c a, ZZ_c p, ZZ_c A, ZZ_c P)

    long ZZ_ReconstructRational "ReconstructRational"(ZZ_c a, ZZ_c b, ZZ_c x, ZZ_c m, ZZ_c a_bound, ZZ_c b_bound)

    void ZZ_GenPrime "GenPrime"(ZZ_c n, long l, long err)
    void ZZ_GenGermainPrime "GenGermainPrime"(ZZ_c n, long l, long err)
    long ZZ_ProbPrime "ProbPrime"(ZZ_c n, long NumTrials)
    void ZZ_RandomPrime "RandomPrime"(ZZ_c n, long l, long NumTrials)
    void ZZ_NextPrime "NextPrime"(ZZ_c n, ZZ_c m, long NumTrials)
    long ZZ_MillerWitness "MillerWitness"(ZZ_c n, ZZ_c w)

    void ZZ_power "NTL::power"(ZZ_c t, ZZ_c x, long e)
    void ZZ_power2 "power2"(ZZ_c x, long e)

    void ZZ_SqrRoot "SqrRoot"(ZZ_c x, ZZ_c a)

    long ZZ_Jacobi "Jacobi"(ZZ_c a, ZZ_c n)
    void ZZ_SqrRootMod "SqrRootMod"(ZZ_c x, ZZ_c a, ZZ_c n)


cdef extern from "ntlwrap_impl.h":
    cdef int ZZ_to_int(ZZ_c* x)
    long ZZ_remove(ZZ_c x, ZZ_c a, ZZ_c p)  # like mpz_remove.
